// (C) 1999, C. Bond, all rights reserved.
// This recursive descent parser is used to evaluate
// mathematical expressions. It has provisions for creating
// variables, assigning values to them and using them in
// subsequent expressions. It also provides for predefined
// constants. The code is based on similar codes by Herbert
// Schildt and Mark Morley, but with additional features and
// a number of limitations eased.
//
// Modifications to the parser are primarily concerned with
// handling return values, which are byte arrays. Since
// these values cannot be returned directly through ordinary
// function calls, pointers are passed instead of values.
// Some changes to the token parser were required, as well
// as slight changes in various descent levels.
//
// This version supports BCD math functions on 32-byte quantities.

#pragma hdrstop
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "evalx.h"
#include "fpconv.h"
#include "fltpt.h"

void gamma(bcd32 x,bcd32 y);
void factorial(bcd32 x,bcd32 y);

static int initflag = 0;

//
// PROTOTYPES FOR INTERNAL DESCENT LEVELS
//
static void XParse(void);
static int  XLevel1(bcd32 q);
static int  XLevel2(bcd32 q);
static int  XLevel3(bcd32 q);
static int  XLevel4(bcd32 q);
static int  XLevel5(bcd32 q);
static int  XLevel6(bcd32 q);

//
// DECLARATIONS FOR VARIABLES WITH FILE SCOPE
//
int   XERRNUM;                // The error number.
char  XERTOK[XTOKLEN + 1];    // The token that generated the error.
int   XERPOS;                 // The offset from the start of the expression.
char* XERANC;                 // Used to calculate ERPOS.
bcd32 xtmp;
XVARIABLE            XVars[MAXXVARS];      // Array for user-defined variables.
static char*         xexpression;          // Pointer to the user's expression .
static unsigned char xtoken[XTOKLEN + 1];  // Holds the current token.
static int           xtype;                // Type of the current token.
enum cidx{ANSID,PIID,LOG2ID};

//
// Add any user accessible "constants" here...  These are "read-only"
// values that are provided as a convenience to the user.  Their values
// can not be permanently changed, although their names may be temporarily
// attached to user specified values.  The first field is the variable
// name, the second is its value.
//
XVARIABLE XConsts[] =
{
//    name,   value 
   { "ans",     0 },
   { "Pi",      0 },
   { "Log2",    0 },
   { 0,0 }
};

//
// AUXILIARY MATH FUNCTIONS
//
 void factorial(bcd32 x,bcd32 y)
 {
    bcdadd(kp1,x,x);
    bcdint(x,x);
    gamma(x,y);
 }

//
// FUNCTION LIST: User name, Number of args, function
//
// Note that for these BCD quantities, the function
// return is done indirectly through a pointer passed as an
// argument. This pointer is always the last argument, hence
// the called function will have one more argument than the
// user specified function. That is, a call to Sqrt(x) by
// the user is translated to a call to:
//
//               bcdsqrt(bcd32 x, bcd32 y)
//
// by the parser. In this case, 'x' is the argument and 'y'
// is the returned value.
//
// WARNING: Calls which take more than one parameter must
// have the function pointer preceded by n-1 NULLs, because
// separate calling sequences are used for functions with
// 1, 2 or more parameters.
//
XFUNCTION XFuncs[] =
{
//    name, #operands, function to call
   { "Sqrt",    1,    bcdsqrt },
   { "Sin",     1,    bcdsin },
   { "Cos",     1,    bcdcos },
   { "Tan",     1,    bcdtan },
   { "Csc",     1,    bcdcsc },
   { "Sec",     1,    bcdsec },
   { "Cot",     1,    bcdcot },
   { "ASin",    1,    bcdasin },
   { "ACos",    1,    bcdacos },
   { "ATan",    1,    bcdatan },
   { "ATan2",   2,    NULL,bcdatan2 },
   { "ACot",    1,    bcdacot },
   { "ACsc",    1,    bcdacsc },
   { "ASec",    1,    bcdasec },
   { "Sinh",    1,    bcdsinh },
   { "Cosh",    1,    bcdcosh },
   { "Tanh",    1,    bcdtanh },
   { "Csch",    1,    bcdcsch },
   { "Sech",    1,    bcdsech },
   { "Coth",    1,    bcdcoth },
   { "ASinh",   1,    bcdasinh },
   { "ACosh",   1,    bcdacosh },
   { "ATanh",   1,    bcdatanh },
   { "ACsch",   1,    bcdacsch },
   { "ASech",   1,    bcdasech },
   { "ACoth",   1,    bcdacoth },
   { "Log",     1,    bcdlog },
   { "Exp",     1,    bcdexp },
   { "Hypot",   2,    NULL,bcdhypot },
   { "Chs",     1,    bcdchs },
   { "Neg",     1,    bcdneg },
   { "Abs",     1,    bcdabs },
   { "Int",     1,    bcdint },
   { "Frac",    1,    bcdfrac },
   { "Gamma",   1,    gamma },
   { "Fact",    1,    factorial },
   { 0,0,NULL }
};

//
// USER ACCESSIBLE UTILITY ROUTINES
//
//  Get name of constant given its index.
//
char* GetXConst(int index)
{
    return XConsts[index].name;
}

//
// Get name of function given its index.
//
char* GetXFuncs(int index)
{
    return XFuncs[index].name;
}

//
// Erases all user-defined variables from memory. Note that constants
// can not be erased or modified in any way by the user.
//
// Returns nothing.
//
void ClearAllXVars()
{
   int i;

   for( i = 0; i < MAXXVARS; i++ )
   {
      *XVars[i].name = 0;
      bcdclear(XVars[i].value);
   }
}

//
// Erases the user-defined variable that is called 'name' from memory.
// Note that constants are not erased. Note also that if the name
// of a constant has been attached to a variable, erasure will make
// make the constant accessible again.
//
// Returns 1 if the variable was found and erased, or 0 if it didn't
// exist.
//
int ClearXVar( char* name )
{
   int i;

   for( i = 0; i < MAXXVARS; i++ )
      if( *XVars[i].name && ! strcmp( name, XVars[i].name ) )
      {
         *XVars[i].name = 0;
         bcdclear(XVars[i].value);
         return( 1 );
      }
   return( 0 );
}

//
// Looks up the specified variable (or constant) known as 'name' and
// returns its contents in 'value'.
//
// First the user-defined variables are searched, then the constants are
// searched.
//
// Returns 1 if the value was found, or 0 if it wasn't.
//
int GetXValue( char* name, bcd32 value )
{
   int i;

// First check the user-defined variables.
   for( i = 0; i < MAXXVARS; i++ )
      if( *XVars[i].name && ! strcmp( name, XVars[i].name ) )
      {
         bcdmov(XVars[i].value,value);
         return( 1 );
      }

// Now check the programmer-defined constants.
   for( i = 0; *XConsts[i].name; i++ )
      if( *XConsts[i].name && ! strcmp( name, XConsts[i].name ) )
      {
         bcdmov(XConsts[i].value,value);
         return( 1 );
      }
   return( 0 );
}

//
// First, it erases any user-defined variable that is called 'name'.  Then
// it creates a new variable called 'name' and gives it the value 'value'.
//
// Returns 1 if the value was added, or 0 if there was no more room.
//
int SetXValue( char *name, bcd32 value )
{
   char b[30];
   int  i;

   ClearXVar( name );
   for( i = 0; i < MAXXVARS; i++ )
      if( ! *XVars[i].name )
      {
         strcpy( XVars[i].name, name );
         XVars[i].name[XVARLEN] = 0;
         bcdmov(value,XVars[i].value);
         return( 1 );
      }
   return( 0 );
}

//
// XERR handler.
//
void XERR(int n) {
    XERRNUM = n;
    XERPOS = xexpression - XERANC - 1;
    strcpy(XERTOK,(char *)xtoken);
}

//
// General purpose routine to check for presence of character
// in a set of characters. Returns 1 if the character 'c' is
// present in the string 'e', otherwise returns 0.
//
int is_in(const char *e,char c)
{
    int i,n;
    n = strlen(e);
    for (i=0;i<n;i++){
        if (c == e[i]) return 1;
    }
    return 0;
}
//
// This function is used to grab the next token from the expression that
// is being evaluated.
//
static void XParse()
{
   char *t, *endptr;
   double value;
   int ccount;

   xtype = 0;
   t = xtoken;
   while( iswhite( *xexpression ) )
      xexpression++;
   if (is_in("+-*/%^(),+!",*xexpression))
   {
      xtype = XDEL;
      *t++ = *xexpression++;
   }
   else if( isnumer( *xexpression ) )
   {
      xtype = XNUM;
      while( isnumer( *xexpression ) )  
        *t++ = *xexpression++;
      if (*xexpression == 'E'  || *xexpression == 'e') {
        *t++ = *xexpression++;
        if (*xexpression == '+' || *xexpression == '-')
           *t++ = *xexpression++;
        while( isexpnum( *xexpression ) )
           *t++ = *xexpression++;
      }
   }
   else if( isalnum( *xexpression )) /* Note: only alpha will reach here */
   {
      xtype = XVAR;
      while( isalnum( *xexpression ) || (*xexpression == '_') )
        *t++ = *xexpression++;
      xtoken[XVARLEN] = 0;
   }
   else if( *xexpression )
   {
      *t++ = *xexpression++;
      *t = 0;
      XERR( E_SYNTAX );
   }
   *t = 0;
   while( iswhite( *xexpression ) )
      xexpression++;
}

//
// XLevel1( bcd32 r )   Internal use only.
//
// This function handles any variable assignment operations.
//
static int XLevel1(bcd32 r)
{
   char t[XVARLEN + 1];

   if( xtype == XVAR )
      if( *xexpression == '=' )
      {
         strcpy( t, (char *)xtoken );
         XParse();
         XParse();
         if( !*xtoken )
         {
            ClearXVar( t );
            return (E_OK);
         }
         XERRNUM = XLevel2( r );
     if (XERRNUM != E_OK) return(XERRNUM); 
         if( ! SetXValue( t, r ) )
         return (E_MAXVARS);
         else return(E_OK);
      }
   XERRNUM = XLevel2(r);
   return XERRNUM;
}

//
// XLevel2( bcd32 r )    Internal use only.
//
// This function handles any addition and subtraction operations.
//
static int XLevel2( bcd32 r )
{
   bcd32 t,z;
   char o;

   bcdclear(t);
   bcdclear(z);
   XERRNUM = XLevel3( r );
   if (XERRNUM != E_OK) return(XERRNUM);
   while( (o = *xtoken) == '+' || o == '-' )
   {
      XParse();
      XERRNUM = XLevel3( t );
      if (XERRNUM != E_OK) return(XERRNUM);
      if( o == '+' ) {
        bcdadd(r,t,z);
        bcdmov(z,r);
      }
      else if( o == '-' ) {
         bcdsub(r,t,z);
         bcdmov(z,r);
        }
   }
   return(E_OK);
}

//
// XLevel3( bcd32 r )    Internal use only.
//
// This function handles any multiplication, division, or modulo.
//
static int XLevel3( bcd32 r )
{
   bcd32 t,z;
   char o;
   int flag;

   bcdclear(t);
   bcdclear(z);
   XERRNUM = XLevel4( r );
   if (XERRNUM != E_OK) return(XERRNUM);
   while( (o = *xtoken) == '*' || o == '/' || o == '%' )
   {
      XParse();
      XERRNUM = XLevel4( t );
      if (XERRNUM != E_OK) return(XERRNUM);
      if( o == '*' ){
        bcdmul(r,t,z);
        bcdmov(z,r);
      }
      else if( o == '/' )
      {
         flag = bcdcmp(t,k0);
         if( flag == 0 ) {
            bcdclear(r);
            return( E_DIVZERO );
         }
         bcddiv(r,t,z);
         bcdmov(z,r);
      }
   }
   return(E_OK);
}

//
// XLevel4( bcd32 r )   Internal use only.
//
// This function handles any "to the power of" operations.
//
static int XLevel4( bcd32 r )
{
   bcd32 t,z;

   XERRNUM = XLevel5( r );
   if (XERRNUM != E_OK) {
    return(XERRNUM);
   }
   while ( *xtoken == '^' )
   {
      XParse();
      XERRNUM = XLevel5( t );
      if (XERRNUM != E_OK) return(XERRNUM);
      bcdpow(r,t,z);
      bcdmov(z,r);
   }
// test for factorial.
    if (*xtoken == '!') {
        factorial(r,z);
        bcdmov(z,r);
        XParse();
        return (E_OK);
    }
// end test for factorial.

   return(E_OK);
}

//
// XLevel5( bcd32 r )    Internal use only.
//
// This function handles any unary + or - signs.
//
static int XLevel5( bcd32 r )
{
   char o = 0;

   if( *xtoken == '+' || *xtoken == '-' )
   {
      o = *xtoken;
      XParse();
   }
   XERRNUM = XLevel6( r );
   if (XERRNUM != E_OK) return(XERRNUM);
   if( o == '-' )
      bcdneg(r,r);
   return(E_OK);
}

//
//  XLevel6( bcd32 r )    Internal use only.
//
// This function handles any literal numbers, variables, or functions.
//
static int XLevel6( bcd32 r )
{
   int  i;
   int  n;
   bcd32 a[4];

   if( *xtoken == '(' )
   {
      XParse();
      if( *xtoken == ')' )
         return( E_NOARG );
      XERRNUM = XLevel1( r );
      if (XERRNUM != E_OK) return (XERRNUM);
      if( *xtoken != ')' )
         return( E_UNBALAN );
      XParse();
   }
   else
   {
      if( xtype == XNUM )
      {
         str2cb2(xtoken,r);
         XParse();
      }
      else if( xtype == XVAR )
      {
         if( *xexpression == '(' )
         {
            for( i = 0; XFuncs[i].args; i++ )
               if( ! strcmp( (char *)xtoken, XFuncs[i].name ) )
               {
                  XParse();
                  n = 0;
                  do
                  {
                     XParse();
                     if( *xtoken == ')' || *xtoken == ',' )
                        return( E_NOARG );
                     bcdclear(a[n]);
                     XERRNUM = XLevel1( a[n] );
                     if(XERRNUM != E_OK) return(XERRNUM);
                     n++;
                  } while( n < 4 && *xtoken == ',' );
          if ( *xtoken != ')' ) return( E_UNBALAN );
                  XParse();
                  if( n != XFuncs[i].args )
                  {
                     strcpy( (char *)xtoken, XFuncs[i].name );
                     return( E_NUMARGS );
                  }
                  if (n == 1) {
                    XFuncs[i].func1( a[0], a[1] );
                  }
                  else if (n == 2) {
                    XFuncs[i].func2( a[0], a[1], a[2] );
                  }
                  bcdmov(a[n],r);
                  return( E_OK );
               }
               if( ! XFuncs[i].name )
                  return( E_BADFUNC );
            }
            else if( ! GetXValue( xtoken, r ) ) {
               return( E_UNKNOWN );
            }
         XParse();
      }
      else
         return( E_SYNTAX );
   }
   return( E_OK );
}

//
// EvaluateX( char* e, bcd32 result, int* epos )
//
// This function is called to evaluate the expression E and return the
// answer in RESULT.
//
// Returns E_OK if the expression is valid, or an error code.
//
int EvaluateX( char *e, bcd32 result, int *epos )
{
// initialize  values.
   if (!initflag) {
   bcdmov(kpi,XConsts[PIID].value);
   bcdmov(kln2,XConsts[LOG2ID].value);
   initflag = 1;
   }
   XERRNUM = E_OK;
   bcdclear(result);
   xexpression = e;
   XERANC = e;

   XParse();
   if( ! *xtoken ) {
      if ( *xexpression == '.' ) {
    *epos = xexpression - XERANC;
        bcdclear(result);
      	return( E_SYNTAX );
      }
      else {
        *epos = xexpression - XERANC - 1;
        bcdclear(result);
      	return( E_EMPTY );
      }
   }
   XERRNUM = XLevel1( result );
   if (XERRNUM != E_OK) {
    *epos = xexpression - XERANC - 1;
        bcdclear(result);
    return(XERRNUM);
   }
   if (*xexpression || *xtoken) {
    *epos = xexpression - XERANC - 1;
        bcdclear(result);
    if (*xtoken == ')' || *xtoken == '(')
		return( E_UNBALAN);
        return( E_SYNTAX );
   }
   bcdmov(result,XConsts[ANSID].value);
   return( E_OK );
}
